import torch
from diffusers import StableDiffusionPipeline, DDIMScheduler, AutoencoderKL
from transformers import AutoFeatureExtractor
from ip_adapter.ip_adapter_faceid import IPAdapterFaceID, IPAdapterFaceIDPlus
from huggingface_hub import hf_hub_download
from insightface.app import FaceAnalysis
from insightface.utils import face_align
import gradio as gr
import cv2
import os

# Generalized model paths for compatibility
base_model_path = "SG161222/Realistic_Vision_V4.0_noVAE"
vae_model_path = "stabilityai/sd-vae-ft-mse"
image_encoder_path = "laion/CLIP-ViT-H-14-laion2B-s32B-b79K"
ip_ckpt = hf_hub_download(repo_id="h94/IP-Adapter-FaceID", filename="ip-adapter-faceid_sd15.bin", repo_type="model")
ip_plus_ckpt = hf_hub_download(repo_id="h94/IP-Adapter-FaceID", filename="ip-adapter-faceid-plusv2_sd15.bin", repo_type="model")

device = "cuda" if torch.cuda.is_available() else "cpu"

# Initialize the noise scheduler
noise_scheduler = DDIMScheduler(
    num_train_timesteps=1000,
    beta_start=0.00085,
    beta_end=0.012,
    beta_schedule="scaled_linear",
    clip_sample=False,
    set_alpha_to_one=False,
    steps_offset=1,
)

# Load models
vae = AutoencoderKL.from_pretrained(vae_model_path).to(dtype=torch.float16 if device == "cuda" else torch.float32)
pipe = StableDiffusionPipeline.from_pretrained(
    base_model_path,
    torch_dtype=torch.float16 if device == "cuda" else torch.float32,
    scheduler=noise_scheduler,
    vae=vae
).to(device)

ip_model = IPAdapterFaceID(pipe, ip_ckpt, device)
ip_model_plus = IPAdapterFaceIDPlus(pipe, image_encoder_path, ip_plus_ckpt, device)

# Initialize FaceAnalysis
app = FaceAnalysis(name="buffalo_l", providers=["CUDAExecutionProvider" if device == "cuda" else "CPUExecutionProvider"])
app.prepare(ctx_id=0, det_size=(640, 640))

STYLE_PRESETS = [
    {
        "title": "Mona Lisa",
        "prompt": "A mesmerizing portrait in the style of Leonardo da Vinci's Mona Lisa, renaissance oil painting, soft sfumato technique, mysterious smile, Florentine background, museum quality, masterpiece"
    },
    {
        "title": "Iron Hero",
        "prompt": "Hyper realistic portrait as a high-tech superhero, wearing advanced metallic suit, arc reactor glow, inside high-tech lab, dramatic lighting, cinematic composition"
    },
    {
        "title": "Ancient Egyptian",
        "prompt": "Portrait as an ancient Egyptian pharaoh, wearing golden headdress and royal regalia, hieroglyphics background, dramatic desert lighting, archaeological discovery style"
    },
    {
        "title": "Sherlock Holmes",
        "prompt": "Victorian era detective portrait, wearing deerstalker hat and cape, holding magnifying glass, foggy London background, mysterious atmosphere, detailed illustration"
    },
    {
        "title": "Star Wars Jedi",
        "prompt": "Epic portrait as a Jedi Master, wearing traditional robes, holding lightsaber, temple background, force aura effect, cinematic lighting, movie poster quality"
    },
    {
        "title": "Van Gogh Style",
        "prompt": "Self-portrait in the style of Vincent van Gogh, bold brushstrokes, vibrant colors, post-impressionist style, emotional intensity, starry background"
    },
    {
        "title": "Greek God",
        "prompt": "Mythological portrait as an Olympian deity, wearing flowing robes, golden laurel wreath, Mount Olympus background, godly aura, classical Greek art style"
    },
    {
        "title": "Medieval Knight",
        "prompt": "Noble knight portrait, wearing ornate plate armor, holding sword and shield, castle background, heraldic designs, medieval manuscript style"
    },
    {
        "title": "Matrix Hero",
        "prompt": "Cyberpunk portrait in digital reality, wearing black trench coat and sunglasses, green code rain effect, dystopian atmosphere, cinematic style"
    },
    {
        "title": "Pirate Captain",
        "prompt": "Swashbuckling pirate captain portrait, wearing tricorn hat and colonial coat, ship's deck background, dramatic sea storm, golden age of piracy style"
    }
]

css = '''
.container {
    background-color: #333333;
    border-radius: 10px;
    padding: 20px;
    color: white;
    margin: auto;
}
.header {
    text-align: center;
    padding: 10px;
    margin-bottom: 20px;
    background: linear-gradient(90deg, #2C3E50, #444444);
    border-radius: 10px;
}
.gallery {
    float: right;
    margin-left: 20px;
    width: 100%;
    height: 800px; /* Make gallery taller */
    display: grid;
    grid-template-columns: repeat(1, 1fr); /* Ensure images are larger */
    grid-gap: 20px; /* Space between images */
    background-color: #333333; /* Apply dark gray color */
    overflow: auto;
}
.gallery img {
    max-width: 100%;
    max-height: 100%;
    object-fit: contain; /* Contain images within their parent container */
}
.dropdown, .textbox, .button {
    margin-bottom: 15px;
    background-color: #444444;
    color: white;
    border-radius: 5px;
}
.textbox {
    height: 550px; /* Adjusted for a larger visible area */
    overflow: auto; /* Add scroll bar for large text */
    background-color: #333333; /* Apply dark gray color */
    border: 2px solid #555555; /* Darkened border for input box */
}
.button {
    background-color: #444444; /* Restore button border */
    color: #FFA500; /* Change text to orange */
    border: 1px solid #FFA500;
    padding: 10px 20px;
    font-size: 16px;
    border-radius: 5px;
}
.button:hover {
    background-color: #555555;
    color: #FFA500;
}
'''

def generate_image(images, gender, style, prompt, progress=gr.Progress(track_tqdm=True)):
    if style and style != "None":
        preset_prompt = next((preset['prompt'] for preset in STYLE_PRESETS if preset['title'] == style), None)
        if preset_prompt:
            prompt = preset_prompt  # Update Custom Prompt with the selected style

    prompt = f"{prompt}, single person, solo portrait, one person only, centered composition"
    negative_prompt = "multiple people, group photo, crowd, double portrait, triple portrait"

    images_embeddings = []
    first_face_image = None

    for image in images:
        img = cv2.imread(image)
        faces = app.get(img)

        if len(faces) > 0:
            embedding = torch.from_numpy(faces[0].normed_embedding).unsqueeze(0)
            images_embeddings.append(embedding)
            if first_face_image is None:
                first_face_image = face_align.norm_crop(img, landmark=faces[0].kps, image_size=224)

    if len(images_embeddings) == 0:
        raise ValueError("No valid faces detected in uploaded images.")

    average_embedding = torch.mean(torch.stack(images_embeddings), dim=0)

    return ip_model_plus.generate(
        prompt=prompt,
        negative_prompt=negative_prompt,
        faceid_embeds=average_embedding,
        face_image=first_face_image,
        width=512,
        height=768,
        num_inference_steps=50,
        guidance_scale=7.5
    )

with gr.Blocks(css=css) as demo:
    with gr.Row():
        with gr.Column(scale=2, elem_classes="container"):
            gr.Markdown("# ✨ Magic Face", elem_classes="header")
            images = gr.Files(file_types=["image"], label="Upload Your Photos")
            gender = gr.Radio(label="Gender", choices=["Female", "Male"], value="Female")
            style = gr.Dropdown(choices=["None"] + [preset['title'] for preset in STYLE_PRESETS], label="Custom Style", value="None")
            prompt = gr.Textbox(label="Custom Prompt", elem_classes="textbox")
            generate_btn = gr.Button("Generate Magic", elem_classes="button")

        with gr.Column(scale=1):
            output_gallery = gr.Gallery(label="Magic Gallery", elem_classes="gallery")

    def update_prompt_from_style(style):
        preset_prompt = next((preset['prompt'] for preset in STYLE_PRESETS if preset['title'] == style), None)
        return {"value": preset_prompt} if preset_prompt else ""

    style.change(
        update_prompt_from_style,
        inputs=[style],
        outputs=[prompt]
    )

    generate_btn.click(
        generate_image,
        inputs=[images, gender, style, prompt],
        outputs=[output_gallery]
    )

demo.launch()
